@@ -26,17 +26,12 @@ mod traits;
 pub mod util;
 
 pub use crate::errors::*;
-pub use crate::memory::{MemoryDatastore, MemoryTransaction};
+pub use crate::memory::MemoryDatastore;
 pub use crate::models::*;
 pub use crate::traits::*;
 
 #[cfg(feature = "rocksdb-datastore")]
 mod rdb;
-#[cfg(feature = "sled-datastore")]
-mod sledds;
 
 #[cfg(feature = "rocksdb-datastore")]
-pub use crate::rdb::{RocksdbDatastore, RocksdbTransaction};
-
-#[cfg(feature = "sled-datastore")]
-pub use crate::sledds::{SledConfig, SledDatastore, SledTransaction};
+pub use crate::rdb::RocksdbDatastore;
