@@ -5,7 +5,7 @@
 
 mod datastore;
 
-pub use self::datastore::{MemoryDatastore, MemoryTransaction};
+pub use self::datastore::MemoryDatastore;
 
 #[cfg(feature = "bench-suite")]
 full_bench_impl!(MemoryDatastore::default());
@@ -17,24 +17,24 @@ full_test_impl!(MemoryDatastore::default());
 #[test]
 fn should_serialize() {
     use super::MemoryDatastore;
-    use crate::{Datastore, SpecificVertexQuery, Transaction, Type};
+    use crate::{Datastore, Identifier, SpecificVertexQuery};
     use tempfile::NamedTempFile;
 
     let path = NamedTempFile::new().unwrap();
 
     let id = {
         let datastore = MemoryDatastore::create(path.path()).unwrap();
-        let trans = datastore.transaction().unwrap();
-        let id = trans.create_vertex_from_type(Type::default()).unwrap();
+        let id = datastore.create_vertex_from_type(Identifier::default()).unwrap();
         datastore.sync().unwrap();
         id
     };
 
     let datastore = MemoryDatastore::read(path.path()).unwrap();
-    let trans = datastore.transaction().unwrap();
-    assert_eq!(trans.get_vertex_count().unwrap(), 1);
-    let vertices = trans.get_vertices(SpecificVertexQuery::new(vec![id])).unwrap();
+    assert_eq!(datastore.get_vertex_count().unwrap(), 1);
+    let vertices = datastore
+        .get_vertices(SpecificVertexQuery::new(vec![id]).into())
+        .unwrap();
     assert_eq!(vertices.len(), 1);
     assert_eq!(vertices[0].id, id);
-    assert_eq!(vertices[0].t, Type::default());
+    assert_eq!(vertices[0].t, Identifier::default());
 }
