@@ -1,13 +1,15 @@
 mod bulk_insert;
 mod edges;
+mod identifiers;
+mod json;
 mod properties;
 mod queries;
-mod types;
 mod vertices;
 
 pub use self::bulk_insert::BulkInsertItem;
 pub use self::edges::{Edge, EdgeKey};
+pub use self::identifiers::Identifier;
+pub use self::json::Json;
 pub use self::properties::{EdgeProperties, EdgeProperty, NamedProperty, VertexProperties, VertexProperty};
 pub use self::queries::*;
-pub use self::types::Type;
 pub use self::vertices::Vertex;
